/*
 * Copyright (C) 2007 Nicolas Joseph
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "gtkcompletion.h"

#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#define GTK_COMPLETION_OBJECT_KEY "gtk_completion"

#define GTK_COMPLETION_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), \
                                       GTK_TYPE_COMPLETION,              \
                                       GtkCompletionPrivate))

typedef struct _GtkCompletionPrivate GtkCompletionPrivate;

struct _GtkCompletionPrivate
{
  GtkTextView *text_view;
};

static void gtk_completion_interface_init (GTypeInstance *instance,
                                           gpointer g_class)
{
  (void)instance;
  (void)g_class;
}

G_DEFINE_TYPE_EXTENDED (GtkCompletion,
                        gtk_completion,
                        GTK_TYPE_COMPLETION_WINDOW,
                        0,
                        G_IMPLEMENT_INTERFACE (GTK_TYPE_BUILDABLE,
                                               gtk_completion_interface_init))

static gchar *get_current_word (GtkCompletion *this)
{
  gchar *word = NULL;
  GtkTextBuffer *buffer = NULL;
	GtkTextMark *mark = NULL;
  GtkTextIter start;
  GtkTextIter end;
  GtkCompletionPrivate *priv = NULL;

  g_return_val_if_fail (this != NULL, word);
  g_return_val_if_fail (GTK_IS_COMPLETION (this), word);

  priv = GTK_COMPLETION_GET_PRIVATE (this);
  buffer = gtk_text_view_get_buffer (priv->text_view);
	mark = gtk_text_buffer_get_insert (buffer);
	gtk_text_buffer_get_iter_at_mark (buffer, &start, mark);
	end = start;
	gtk_text_iter_backward_word_start (&start);
	gtk_text_iter_forward_word_end (&end);
  return gtk_text_iter_get_slice (&start, &end);
}

static void completion_activate (GtkWidget *widget, const gchar *select,
                                 GtkCompletion *this)
{
  gchar *current = NULL;
  GtkTextMark *mark = NULL;
  GtkTextIter text_iter;
  GtkTextBuffer *buffer = NULL;
  GtkCompletionPrivate *priv = NULL;

  g_return_if_fail (this != NULL);
  g_return_if_fail (GTK_IS_COMPLETION (this));
  g_return_if_fail (select != NULL);

  priv = GTK_COMPLETION_GET_PRIVATE (this);
  buffer = gtk_text_view_get_buffer (priv->text_view);
  mark = gtk_text_buffer_get_insert (buffer);
  gtk_text_buffer_get_iter_at_mark (buffer, &text_iter, mark);
  current = get_current_word (this);
  if (current != NULL)
  {
    select += strlen (current);
    g_free (current), current = NULL;
  }
  gtk_text_buffer_insert (buffer, &text_iter, select, -1);

  (void)widget;
}

static void gtk_completion_popup (GtkCompletion *this)
{
  g_return_if_fail (this != NULL);
  g_return_if_fail (GTK_IS_COMPLETION (this));

  /* Repositionnement de la popup */
  {
    gint x = -1;
    gint y = -1;
    gint xor = -1;
    gint yor = -1;
    GdkRectangle location;
    GtkTextBuffer *buffer = NULL;
    GtkTextMark *mark = NULL;
    GtkTextIter iter;
    GdkWindow *win = NULL;
    GtkCompletionPrivate *priv = NULL;

    priv = GTK_COMPLETION_GET_PRIVATE (this);
    buffer = gtk_text_view_get_buffer (priv->text_view);
    mark = gtk_text_buffer_get_insert (buffer);
    gtk_text_buffer_get_iter_at_mark (buffer, &iter, mark);
    gtk_text_view_get_iter_location (priv->text_view, &iter, &location);

    win = gtk_text_view_get_window (priv->text_view, GTK_TEXT_WINDOW_TEXT);

    gtk_text_view_buffer_to_window_coords (priv->text_view,
                                           GTK_TEXT_WINDOW_TEXT,
                                           location.x + location.width,
                                           location.y, &x, &y);
    gdk_window_get_origin (win, &xor, &yor);
    gtk_completion_window_move (GTK_COMPLETION_WINDOW (this),
                                x + xor, y + yor);
  }
  {
    gchar *current = NULL;

    current = get_current_word (this);
    gtk_completion_window_set_prefix (GTK_COMPLETION_WINDOW (this), current);
    g_free (current), current = NULL;
  }
  g_signal_connect (G_OBJECT (this), "completion-activate",
                    G_CALLBACK (completion_activate), this);
  gtk_widget_show_all (GTK_WIDGET (this));
}

static gboolean gtk_completion_key_press_event (GtkWidget *widget,
                                                GdkEventKey *event,
                                                gpointer data)
{
  gboolean ret = FALSE;
  GtkCompletion *this = data;

  g_return_val_if_fail (this != NULL, ret);
  g_return_val_if_fail (GTK_IS_COMPLETION (this), ret);

	if (event->keyval == GDK_Return && event->state & GDK_CONTROL_MASK)
	{
    gtk_completion_popup (this);
    ret = TRUE;
	}
	else
	{
	  ret = GTK_WIDGET_CLASS (GTK_TEXT_VIEW_GET_CLASS (widget))->key_press_event (widget, event);
	}
	return ret;
}

static void gtk_completion_finalize (GObject *object)
{
  GtkCompletion *this  = GTK_COMPLETION (object);
  GtkCompletionPrivate *priv = NULL;

  g_return_if_fail (this != NULL);
  g_return_if_fail (GTK_IS_COMPLETION (this));

  priv = GTK_COMPLETION_GET_PRIVATE (this);
  g_signal_handlers_disconnect_matched (priv->text_view, G_SIGNAL_MATCH_DATA,
                                        0, 0, NULL, NULL, this);
  G_OBJECT_CLASS (gtk_completion_parent_class)->finalize (object);
}

static void gtk_completion_class_init (GtkCompletionClass *klass)
{
  GObjectClass *gobject_class = NULL;

  g_return_if_fail (klass != NULL);

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = gtk_completion_finalize;

  g_type_class_add_private (klass, sizeof (GtkCompletionPrivate));
}

static void gtk_completion_init (GtkCompletion *this)
{
  g_return_if_fail (this != NULL);
}

/**
 * gtk_completion_new:
 * @items: The items to show in completion popup.
 *
 * Create a new #GtkCompletion object with items @items.
 *
 * Returns: a new #GtkCompletion object, or %NULL on error.
 */
GtkCompletion *gtk_completion_new (const gchar **items)
{
  GtkCompletion *this = NULL;
  GtkCompletionPrivate *priv = NULL;

  this = g_object_new (GTK_TYPE_COMPLETION, NULL);
  priv = GTK_COMPLETION_GET_PRIVATE (this);
  priv->text_view = NULL;
  if (items != NULL)
  {
    gint i;
    GList *l = NULL;

    for (i = 0; items[i] != NULL; i++)
    {
      l = g_list_append (l, g_strdup (items[i]));
    }
    gtk_completion_window_set_items (GTK_COMPLETION_WINDOW (this), l);
    g_list_foreach (l, (GFunc)g_free, NULL);
    g_list_free (l), l = NULL;
  }
  return this;
}

/**
 * gtk_completion_attach:
 * @this: A #GtkCompletion.
 * @text_view: The #GtkTextView to attach to.
 *
 * Attache a #GtkCompletion to @text_view.
 */
void gtk_completion_attach (GtkCompletion *this, GtkTextView *text_view)
{
  g_return_if_fail (this != NULL);
  g_return_if_fail (GTK_IS_COMPLETION (this));
  g_return_if_fail (text_view != NULL);
  g_return_if_fail (GTK_IS_TEXT_VIEW (text_view));

	if (g_object_get_data(G_OBJECT (text_view), GTK_COMPLETION_OBJECT_KEY) == NULL)
	{
    GtkCompletionPrivate *priv = NULL;

    priv = GTK_COMPLETION_GET_PRIVATE (this);
    priv->text_view = text_view;
    g_object_set_data(G_OBJECT (priv->text_view), GTK_COMPLETION_OBJECT_KEY,
                      this);
    g_signal_connect (G_OBJECT (priv->text_view), "key-press-event",
                      G_CALLBACK (gtk_completion_key_press_event), this);
    g_signal_connect_swapped (G_OBJECT (priv->text_view), "destroy",
                              G_CALLBACK (gtk_widget_destroy), this);
	}
}

/**
 * gtk_completion_new_attach:
 * @view: The #GtkTextView to attach to.
 * @text_view: The items to show in completion popup.
 *
 * Create a new #GtkCompletion object attached to @text_view with items @items.
 *
 * Returns: a new #GtkCompletion object, or %NULL on error.
 */
GtkCompletion *gtk_completion_new_attach (GtkTextView *text_view,
                                          const gchar **items)
{
  GtkCompletion *this = NULL;

  g_return_val_if_fail (text_view != NULL, this);
  g_return_val_if_fail (GTK_IS_TEXT_VIEW (text_view), this);

  this = gtk_completion_new (items);
  gtk_completion_attach (this, text_view);
  return this;
}

/**
 * gtk_completion_get_from_text_view:
 * @text_view: A #GtkTextView.
 *
 * Retrieves the #GtkTextView object attached to a text view.
 *
 * Returns: the #GtkTextView object, or %NULL if there is no #GtkSpell
 * attached to @text_view.
 */
GtkCompletion *gtk_completion_get_from_text_view (GtkTextView *text_view)
{
	return g_object_get_data(G_OBJECT (text_view), GTK_COMPLETION_OBJECT_KEY);
}

/**
 * gtk_completion_detach:
 * @this: A #GtkCompletion.
 *
 * Detaches this #GtkCompletion from its text view.  Use
 * gtk_completion_get_from_text_view() to retrieve a GtkCompletion from a
 * #GtkTextView.
 */
void gtk_completion_detach (GtkCompletion *this)
{
  GtkCompletionPrivate *priv = NULL;

  g_return_if_fail (this != NULL);
  g_return_if_fail (GTK_IS_COMPLETION (this));

  priv = GTK_COMPLETION_GET_PRIVATE (this);
	g_object_set_data(G_OBJECT (priv->text_view), GTK_COMPLETION_OBJECT_KEY,
	                  NULL);
}
