/*
 * Copyright (C) 2007 Nicolas Joseph
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "gtkcompletionwindow.h"

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "gtktreeviewfilter.h"

#define MARGIN 15

enum
{
  COMPLETION_ACTIVATE,
  LAST_SIGNAL
};

#define GTK_COMPLETION_WINDOW_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), \
                                              GTK_TYPE_COMPLETION_WINDOW,       \
                                              GtkCompletionWindowPrivate))

typedef struct _GtkCompletionWindowPrivate GtkCompletionWindowPrivate;

struct _GtkCompletionWindowPrivate
{
  GtkTreeViewFilter *tree_view;
};

static guint completion_window_signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE (GtkCompletionWindow, gtk_completion_window, GTK_TYPE_WINDOW)

static void gtk_completion_row_activated (GtkTreeView *p_tree_view,
                                          GtkTreePath *arg1,
                                          GtkTreeViewColumn *arg2,
                                          GtkCompletionWindow *this)
{
  gchar *str = NULL;
  GtkTreeIter iter;
  GtkTreeModel *p_tree_model = NULL;

  g_return_if_fail (this != NULL);
  g_return_if_fail (GTK_IS_COMPLETION_WINDOW (this));

  p_tree_model = gtk_tree_view_get_model (p_tree_view);
  gtk_tree_model_get_iter (p_tree_model, &iter, arg1);
  gtk_tree_model_get (p_tree_model, &iter, 0, &str, -1);
  g_signal_emit (this, completion_window_signals[COMPLETION_ACTIVATE], 0, str);
  g_free (str), str = NULL;

  gtk_widget_hide (GTK_WIDGET (this));

  (void)arg2;
}

static GtkWindow *get_active_window (void)
{
  GtkWindow *active = NULL;
  GList *windows = NULL;
  GList *w = NULL;

  windows = gtk_window_list_toplevels ();
  for (w = g_list_first (windows); w != NULL; w = g_list_next (w))
  {
    GtkWindow *window = NULL;

    window = GTK_WINDOW (w->data);
    if (gtk_window_is_active (window))
    {
      active = window;
      break;
    }
  }
  return active;
}

static gboolean gtk_completion_window_focus_out_event (GtkWidget *widget,
                                                       GdkEventFocus *event)
{
  g_return_val_if_fail (widget != NULL, TRUE);
  g_return_val_if_fail (GTK_IS_COMPLETION_WINDOW (widget), TRUE);

  gtk_widget_hide (widget);

  (void)event;

  return FALSE;
}

static gboolean gtk_completion_window_key_press_event (GtkWidget *widget,
                                                       GdkEventKey *event)
{
  gboolean ret = FALSE;

  g_return_val_if_fail (widget != NULL, ret);
  g_return_val_if_fail (GTK_IS_COMPLETION_WINDOW (widget), ret);

	if (event->keyval == GDK_Escape)
	{
	  gtk_widget_hide (widget);
	}
	else
	{
	  ret = GTK_WIDGET_CLASS (gtk_completion_window_parent_class)->key_press_event (widget, event);
	}
	return ret;
}

static void gtk_completion_window_class_init (GtkCompletionWindowClass *klass)
{
  GtkWidgetClass *widget_class = NULL;

  g_return_if_fail (klass != NULL);

  widget_class = GTK_WIDGET_CLASS (klass);

  widget_class->focus_out_event = gtk_completion_window_focus_out_event;
  widget_class->key_press_event = gtk_completion_window_key_press_event;

  completion_window_signals[COMPLETION_ACTIVATE] =
    g_signal_new ("completion-activate",
                  G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (GtkCompletionWindowClass, completion_activate),
                  NULL, NULL,
                  g_cclosure_marshal_VOID__STRING,
                  G_TYPE_NONE, 1,
                  G_TYPE_STRING);
  g_type_class_add_private (klass, sizeof (GtkCompletionWindowPrivate));
}

static void gtk_completion_window_init (GtkCompletionWindow *this)
{
  GtkCompletionWindowPrivate *priv = NULL;

  g_return_if_fail (this != NULL);

  gtk_window_set_keep_above (GTK_WINDOW (this), TRUE);
  gtk_window_set_decorated (GTK_WINDOW (this), FALSE);
  gtk_window_set_skip_taskbar_hint (GTK_WINDOW (this), TRUE);
  gtk_window_set_skip_pager_hint (GTK_WINDOW (this), TRUE);
  {
    GtkWindow *active = NULL;
    GtkWindowGroup *group = NULL;

    active = get_active_window ();
    group = gtk_window_get_group (active);
    gtk_window_group_add_window (group, GTK_WINDOW (this));
    gtk_window_set_transient_for (GTK_WINDOW (this), active);
  }
  gtk_window_set_resizable (GTK_WINDOW (this), FALSE);
  gtk_widget_set_size_request (GTK_WIDGET (this), 200, 300);

  priv = GTK_COMPLETION_WINDOW_GET_PRIVATE (this);
  priv->tree_view = GTK_TREE_VIEW_FILTER (gtk_tree_view_filter_new ());
  g_signal_connect (G_OBJECT (gtk_tree_view_filter_get_tree_view (priv->tree_view)),
                    "row-activated",
                    G_CALLBACK (gtk_completion_row_activated), this);
  gtk_container_add (GTK_CONTAINER (this), GTK_WIDGET (priv->tree_view));
}

GtkWidget *gtk_completion_window_new (void)
{
  return GTK_WIDGET (g_object_new (GTK_TYPE_COMPLETION_WINDOW, NULL));
}

void gtk_completion_window_move (GtkCompletionWindow *this, gint x, gint y)
{
  GdkScreen *screen = NULL;
  gint width = -1;
  gint height = -1;
  gint ch = -1;
  gint cw = -1;

  g_return_if_fail (this != NULL);
  g_return_if_fail (GTK_IS_COMPLETION_WINDOW (this));

  screen = gtk_widget_get_screen (GTK_WIDGET (this));
  width = gdk_screen_get_width (screen);
  height = gdk_screen_get_height (screen);

  gtk_window_get_size (GTK_WINDOW (this), &cw, &ch);

  if (x + cw > width)
  {
    x = width - cw - MARGIN;
  }
  else if (x < MARGIN)
  {
    x = MARGIN;
  }
  if (y + ch > height)
  {
    y = height - ch - MARGIN;
  }
  else if (y < MARGIN)
  {
    y = MARGIN;
  }
  gtk_window_move (GTK_WINDOW (this), x, y);
}

void gtk_completion_window_set_items (GtkCompletionWindow *this, GList *items)
{
  GtkCompletionWindowPrivate *priv = NULL;

  priv = GTK_COMPLETION_WINDOW_GET_PRIVATE (this);
  gtk_tree_view_filter_set_items (priv->tree_view, items);
}

void gtk_completion_window_set_prefix (GtkCompletionWindow *this,
                                      const gchar *prefix)
{
  GtkCompletionWindowPrivate *priv = NULL;

  priv = GTK_COMPLETION_WINDOW_GET_PRIVATE (this);
  gtk_tree_view_filter_set_prefix (priv->tree_view, prefix);
}
