/*
 * Copyright (C) 2007 Nicolas Joseph
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "gtktreeviewfilter.h"

#include <string.h>
#include <gtk/gtk.h>

#define GTK_TREE_VIEW_FILTER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), \
                                             GTK_TYPE_TREE_VIEW_FILTER,        \
                                             GtkTreeViewFilterPrivate))

typedef struct _GtkTreeViewFilterPrivate GtkTreeViewFilterPrivate;

struct _GtkTreeViewFilterPrivate
{
  GtkEntry *entry;
  GtkTreeModelFilter *filter;
  GtkTreeView *tree_view;
  GtkListStore *list_store;
};

G_DEFINE_TYPE (GtkTreeViewFilter, gtk_tree_view_filter, GTK_TYPE_FRAME)

static gboolean filter_func (GtkTreeModel *p_tree_model, GtkTreeIter *iter,
                             gpointer data)
{
  GtkTreeViewFilter *this = data;
  gboolean visible = TRUE;
  gchar *item = NULL;

  g_return_val_if_fail (this != NULL, visible);
  g_return_val_if_fail (GTK_IS_TREE_VIEW_FILTER (this), visible);

  gtk_tree_model_get (p_tree_model, iter, 0, &item, -1);
  if (item != NULL)
  {
    const gchar *select_item = NULL;
    GtkTreeViewFilterPrivate *priv = NULL;

    priv = GTK_TREE_VIEW_FILTER_GET_PRIVATE (this);

    select_item = gtk_entry_get_text (priv->entry);
    if (select_item != NULL && select_item[0] != '\0'
        && strstr (item, select_item) == NULL)
    {
        visible = FALSE;
    }
    g_free (item), item = NULL;
  }
  return visible;
}

static void cb_refresh (GtkTreeViewFilter *this)
{
  GtkTreeViewFilterPrivate *priv = NULL;

  g_return_if_fail (this != NULL);
  g_return_if_fail (GTK_IS_TREE_VIEW_FILTER (this));

  priv = GTK_TREE_VIEW_FILTER_GET_PRIVATE (this);

  gtk_tree_model_filter_refilter (priv->filter);
}

static void gtk_tree_view_filter_class_init (GtkTreeViewFilterClass *klass)
{
  g_return_if_fail (klass != NULL);

  g_type_class_add_private (klass, sizeof (GtkTreeViewFilterPrivate));
}

static void gtk_tree_view_filter_init (GtkTreeViewFilter *this)
{
  GtkWidget *vbox = NULL;

  g_return_if_fail (this != NULL);

  gtk_frame_set_shadow_type (GTK_FRAME (this), GTK_SHADOW_ETCHED_IN);
  vbox = gtk_vbox_new (FALSE, 3);
  gtk_container_add (GTK_CONTAINER (this), vbox);

  {
    GtkWidget *p_hbox = NULL;
    GtkTreeViewFilterPrivate *priv = NULL;

    priv = GTK_TREE_VIEW_FILTER_GET_PRIVATE (this);

    p_hbox = gtk_hbox_new (FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox), GTK_WIDGET (p_hbox), FALSE, FALSE, 0);

    priv->entry = GTK_ENTRY (gtk_entry_new ());
    gtk_box_pack_start (GTK_BOX (p_hbox), GTK_WIDGET (priv->entry), TRUE,
                        TRUE, 0);
    g_signal_connect_swapped (G_OBJECT (priv->entry), "changed",
                              G_CALLBACK (cb_refresh), this);

    {
      GtkWidget *p_scrolled_window = NULL;
      GtkCellRenderer *renderer = NULL;
      GtkTreeViewColumn *column = NULL;

      p_scrolled_window = gtk_scrolled_window_new (NULL, NULL);
      gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (p_scrolled_window),
                                      GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
      gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (p_scrolled_window),
                                           GTK_SHADOW_OUT);
      gtk_box_pack_start (GTK_BOX (vbox), p_scrolled_window, TRUE, TRUE, 0);

      priv->list_store = gtk_list_store_new (1, G_TYPE_STRING);
      priv->filter = GTK_TREE_MODEL_FILTER (gtk_tree_model_filter_new (GTK_TREE_MODEL (priv->list_store), NULL));
      gtk_tree_model_filter_set_visible_func (priv->filter, filter_func, this, NULL);
      priv->tree_view = GTK_TREE_VIEW (gtk_tree_view_new_with_model (GTK_TREE_MODEL (priv->filter)));
      gtk_tree_view_set_headers_visible (priv->tree_view, FALSE);
      renderer = gtk_cell_renderer_text_new ();
      column = gtk_tree_view_column_new_with_attributes ("FilterTreeView",
                                                         renderer, "text",
                                                         0, NULL);
      gtk_tree_view_append_column (priv->tree_view, column);
      gtk_container_add (GTK_CONTAINER (p_scrolled_window),
                         GTK_WIDGET (priv->tree_view));
    }
  }
}

GtkWidget *gtk_tree_view_filter_new (void)
{
  GtkTreeViewFilter *this = NULL;

  this = g_object_new (GTK_TYPE_TREE_VIEW_FILTER, NULL);
  return GTK_WIDGET (this);
}

GtkTreeView *gtk_tree_view_filter_get_tree_view (GtkTreeViewFilter *this)
{
  GtkTreeView *tree_view = NULL;
  GtkTreeViewFilterPrivate *priv = NULL;

  g_return_val_if_fail (this != NULL, tree_view);
  g_return_val_if_fail (GTK_IS_TREE_VIEW_FILTER (this), tree_view);

  priv = GTK_TREE_VIEW_FILTER_GET_PRIVATE (this);
  tree_view = priv->tree_view;
  return tree_view;
}

void gtk_tree_view_filter_set_items (GtkTreeViewFilter *this, GList *items)
{
  GList *l = NULL;
  GtkTreeViewFilterPrivate *priv = NULL;

  g_return_if_fail (this != NULL);
  g_return_if_fail (GTK_IS_TREE_VIEW_FILTER (this));

  priv = GTK_TREE_VIEW_FILTER_GET_PRIVATE (this);
  gtk_list_store_clear (priv->list_store);
  for (l = g_list_first (items); l != NULL; l = g_list_next (l))
  {
    GtkTreeIter iter;

    gtk_list_store_append (priv->list_store, &iter);
    gtk_list_store_set (priv->list_store, &iter, 0, (gchar *)l->data, -1);
  }
  cb_refresh (this);
}

void gtk_tree_view_filter_set_prefix (GtkTreeViewFilter *this,
                                      const gchar *prefix)
{
  GtkTreeViewFilterPrivate *priv = NULL;

  g_return_if_fail (this != NULL);
  g_return_if_fail (GTK_IS_TREE_VIEW_FILTER (this));

  priv = GTK_TREE_VIEW_FILTER_GET_PRIVATE (this);
  gtk_entry_set_text (priv->entry, prefix);
}
